from osiris_skretrieval.osiris_triplet import OsirisTriplet
import numpy as np
from copy import copy
import sasktran as sk
from scipy import optimize
import logging
from skretrieval.core.radianceformat import RadianceBase
import xarray as xr
from scipy import interpolate
from skretrieval.retrieval.tikhonov import two_dim_vertical_first_deriv, two_dim_vertical_second_deriv


class OsirisAerosolRetrieval(OsirisTriplet):
    def __init__(self, psize_clim, extinction_wavelength=750, max_numden=1000,
                 use_apriori_for_upperbound=False,
                 require_all_wavelengths_valid=True,
                 use_jacobian_for_lowerbound=True,
                 use_chi_sq_for_lowerbound=False,
                 double_normalize=True,
                 assume_error_fraction=None,
                 triplet_wavelengths=(470.0, 675.0, 750.0, 805.0),
                 normalization_covariance=False, **kwargs):
        self._normalization_range = [30000, 40000]
        self._base_triplet = [{'wavelengths': triplet_wavelengths,
                               'weights': [1.0, 1.0, 1.0, 1.0],
                               'upper_altitude_m': 40000,
                               'normalization_range': copy(self._normalization_range),
                               'lower_altitude_m': 0}]
        super().__init__(triplets=copy(self._base_triplet), **kwargs)

        self._use_jacobian_for_lowerbound = use_jacobian_for_lowerbound
        self._use_chi_sq_for_lowerbound = use_chi_sq_for_lowerbound

        self._double_normalize = double_normalize

        self._use_apriori_for_upperbound = use_apriori_for_upperbound
        self._require_all_wavelengths_valid = require_all_wavelengths_valid

        self._assume_error_fraction = assume_error_fraction

        self._normalization_covariance = normalization_covariance

        if self._use_apriori_for_upperbound:
            self._initial_aero = self._species.climatology.get_parameter(self._species.species, 0, 0, self._retrieval_altitudes, 54372)
            self._initial_aero[self._initial_aero == 0] = 1e-20

        self._apriori_upperbound = 0
        self._extinction_wavelength= extinction_wavelength
        self._psize_clim = psize_clim
        self._max_numden = max_numden

        self._initial_mode_radius = psize_clim['SKCLIMATOLOGY_LOGNORMAL_MODERADIUS_MICRONS'][0]
        self._initial_mode_width = psize_clim['SKCLIMATOLOGY_LOGNORMAL_MODEWIDTH'][0]

        self._tanalts = kwargs['measurement_l1'].tangent_locations().altitude.values
        self._min_update_alt = 5000
        self._weight_iters = 4
        self._max_alt = 40000

        self._l1_ray = None

        self._ref_range = 10 # Set this small to only use one altitude in the normalization

        self._ref_idx = []

        self._min_fractional_sensitivity = 0.05e-3

        self._max_update_factor = 5

        self._max_norm_range = 10000

    def measurement_vector_optimization(self, l1_data):
        selected_data = l1_data.data.sel(wavelength=self._base_triplet[0]['wavelengths'], method='nearest')
        return np.log(selected_data['radiance'])

    def measurement_vector_unnormalized(self, l1_data):
        self._high_altitude_normalize = False
        y = self.measurement_vector(l1_data)
        self._high_altitude_normalize = True

        return y

    def optimize_settings(self, forward_model, measurement_l1):
        """
        Optimizes the triplet and the normalization altitude by running several radiative transfer calculations
        """
        self._triplets = self._optimal_triplet(forward_model, measurement_l1)

        # Update the minimum altitude from the Jacobian

        # Run the calculation with aer=0 to get the rayleigh background
        old_aerosol = copy(self._species.climatology['SKCLIMATOLOGY_AEROSOL_CM3'])
        self._species.climatology['SKCLIMATOLOGY_AEROSOL_CM3'] *= 0.0
        self._l1_ray = forward_model.calculate_radiance()
        self._l1_ray.data['tangent_altitude'] = self._l1_ray.tangent_locations()['altitude']
        self._l1_ray.data = self._l1_ray.data.swap_dims({'los': 'tangent_altitude'})
        self._species.climatology['SKCLIMATOLOGY_AEROSOL_CM3'] = old_aerosol

        # First find the normalization altitude for each wavelength
        meas_l1_copy = measurement_l1.data.copy()
        meas_l1_copy['tangent_altitude'] = measurement_l1.tangent_locations()['altitude']
        meas_l1_copy = meas_l1_copy.swap_dims({'los': 'tangent_altitude'})

        log_radiance_difference = np.log(meas_l1_copy)['radiance'] - np.log(self._l1_ray.data)['radiance']

        self._triplets[0]['wavelength_normalization_range'] = []
        for widx, w in enumerate(log_radiance_difference.wavelength.values):
            diff = log_radiance_difference.sel(wavelength=w)
            diff = diff.where((diff.tangent_altitude > self._normalization_range[0]) & (diff.tangent_altitude < self._normalization_range[1]), drop=True)

            min_y = diff.min()
            good_alts = diff.where(diff <= min_y + 0.01, drop=True).tangent_altitude.values

            triplet_high_alt_norm = float(good_alts.max() + 10)
            triplet_low_alt_norm = np.nanmax([float(good_alts.min() - 10), triplet_high_alt_norm - self._max_norm_range])

            self._triplets[0]['wavelength_normalization_range'].append([triplet_low_alt_norm, triplet_high_alt_norm])

        # To calculate the normalization we need to let the triplet go up to the max of the normalization range
        self._triplets[0]['upper_altitude_m'] = self._normalization_range[1]
        if not self._retrieve_on_osiris_grid:
            self._retrieval_altitudes = np.arange(self._initial_lower_bound, self._initial_upper_bound, self._ret_spacing_m)
        y_meas = self.measurement_vector_unnormalized(measurement_l1)

        dy = y_meas['y']

        if self._double_normalize:
            possible_ref_idx = (y_meas['y_altitude'] > self._normalization_range[0]) & (y_meas['y_altitude'] < self._normalization_range[1])

            # Only normalize at altitudes where all of the wavelengths are not nan
            nan_mask = np.isnan(log_radiance_difference).any(dim='wavelength').sel(tangent_altitude=y_meas['y_altitude'])
            possible_ref_idx = possible_ref_idx & ~nan_mask.values

            if np.sum(possible_ref_idx) == 0:
                raise ValueError('No possible altitudes to normalize the aerosol retrieval at')

            ref_idx = dy[possible_ref_idx] <= np.min(dy[possible_ref_idx]) + 0.01
            ref_alt = y_meas['y_altitude'][possible_ref_idx][ref_idx]

            upper_altitude = np.max([np.nanmin(ref_alt) - self._ref_range, self._normalization_range[0]])
            self._triplets[0]['normalization_range'] = [upper_altitude, np.nanmax(ref_alt) + 10]
        else:
            min_norm_alt = np.nanmax([a[0] for a in self._triplets[0]['wavelength_normalization_range']])
            upper_altitude = np.max([min_norm_alt - self._ref_range, self._normalization_range[0]])

            # Set the weights above the normalization altitude to 0
            for widx in range(len(self._triplets[0]['wavelengths'])):
                bad_idx = self._triplets[0]['weight_altitudes'] > self._triplets[0]['wavelength_normalization_range'][widx][0] - 10
                self._triplets[0]['weights'][widx, bad_idx] = 0

        self._triplets[0]['upper_altitude_m'] = upper_altitude
        highest_usable_tangent_altitude = self._tanalts[self._tanalts < upper_altitude][-1]

        lower_altitude = self._retrieval_altitudes[0]
        if not self._use_apriori_for_upperbound:
            self._retrieval_altitudes = np.arange(lower_altitude, upper_altitude, self._ret_spacing_m)

            # Adjust the highest altitude to be only just above the highest useable tangent altitude
            self._retrieval_altitudes = self._retrieval_altitudes[self._retrieval_altitudes < highest_usable_tangent_altitude + self._ret_spacing_m / 2.0]
        else:
            self._retrieval_altitudes = self._retrieval_altitudes[self._retrieval_altitudes >= lower_altitude - 5]
            self._apriori_upperbound = highest_usable_tangent_altitude + self._ret_spacing_m / 2.0
            self._initial_aero = self._species.climatology.get_parameter(self._species.species, 0, 0, self._retrieval_altitudes, 54372)
            self._initial_aero[self._initial_aero == 0] = 1e-20

        forward_model.engine.options['wfheights'] = self._retrieval_altitudes
        forward_model.engine.options['wfwidths'] = np.gradient(self._retrieval_altitudes)

    def min_alt_from_jacobian(self, y):
        low_alt_from_y = y['y_altitude'][~np.isnan(y['y'])][0]
        low_alt_from_y = self._retrieval_altitudes[self._retrieval_altitudes > low_alt_from_y - self._ret_spacing_m / 2.0][0]

        K = copy(y['jacobian'])

        if self._state_vector_space == 'log':
            # Convert to linear space as the relative Jacobian does not give a good idea of sensitivity
            x = self.state_vector()
            K /= np.exp(x)[np.newaxis, :]

        altitude_sensitivity = np.nansum(K, axis=0)
        max_arg = np.nanargmax(altitude_sensitivity)

        try:
            lowest_idx = np.nonzero(altitude_sensitivity[:max_arg] < self._min_fractional_sensitivity)[0][-1]
        except IndexError:
            lowest_idx = 0

        low_alt_from_K = self._retrieval_altitudes[lowest_idx]

        return np.nanmax([low_alt_from_K, low_alt_from_y])

    def adjust_min_altitude(self, new_min_altitude):
        if new_min_altitude != self._retrieval_altitudes[0]:
            logging.info('Aerosol retrieval adjusting minimum altitude to {}'.format(new_min_altitude))
            if not self._use_apriori_for_upperbound:
                self._retrieval_altitudes = np.arange(new_min_altitude, self._retrieval_altitudes[-1] + 1, self._ret_spacing_m)
            else:
                self._retrieval_altitudes = self._retrieval_altitudes[self._retrieval_altitudes >= new_min_altitude - 5]

            if self._use_apriori_for_upperbound:
                self._initial_aero = self._species.climatology.get_parameter(self._species.species, 0, 0,
                                                                             self._retrieval_altitudes, 54372)
                self._initial_aero[self._initial_aero == 0] = 1e-20

    def _optimal_triplet(self, forward_model, measurement_l1):
        forward_model.engine.options['calcwf'] = 1
        forward_model._mask_osiris_bad_pixels = False
        K, S = self._normalized_radiance_jacobians(forward_model.calculate_radiance(), measurement_l1)
        forward_model.engine.options['calcwf'] = 2
        forward_model._mask_osiris_bad_pixels = True

        w = self.minimize_w(K, S, use_constraints=False)

        triplet = copy(self._base_triplet)
        triplet[0]['weights'] = w
        triplet[0]['weight_altitudes'] = self._tanalts

        return triplet

    def _set_mode_radius_and_width(self, rg, sg):
        # Get the current cross section
        xs_current = self._species.optical_property.calculate_cross_sections(sk.MSIS90(), 0, 0, 20000, 54372, self._extinction_wavelength).total

        # Set the new parameters
        self._psize_clim['SKCLIMATOLOGY_LOGNORMAL_MODERADIUS_MICRONS'] = np.ones_like(self._psize_clim['SKCLIMATOLOGY_LOGNORMAL_MODERADIUS_MICRONS']) * rg
        self._psize_clim['SKCLIMATOLOGY_LOGNORMAL_MODEWIDTH'] = np.ones_like(self._psize_clim['SKCLIMATOLOGY_LOGNORMAL_MODEWIDTH']) * sg

        # Get the new cross section
        xs_new = self._species.optical_property.calculate_cross_sections(sk.MSIS90(), 0, 0, 20000, 54372, self._extinction_wavelength).total

        # Update the number density
        self._species.climatology['SKCLIMATOLOGY_AEROSOL_CM3'] *= xs_current / xs_new

    def _normalized_radiance_jacobians(self, forward_l1, measurement_l1):
        tangent_locations = forward_l1.tangent_locations()

        ref_alt_idx = np.argmin(np.abs(tangent_locations.altitude.values - 30000))

        forward_radiance = forward_l1.data.sel(wavelength=self._base_triplet[0]['wavelengths'], method='nearest')
        meas_radiance = measurement_l1.data.sel(wavelength=self._base_triplet[0]['wavelengths'], method='nearest').copy(True)

        meas_radiance['radiance_noise'] = 0.01*meas_radiance['radiance']

        meas_norm = meas_radiance.isel(los=ref_alt_idx)

        jacobian = forward_radiance['wf'] / forward_radiance['radiance']

        rad = meas_radiance['radiance'].values
        S_noise = np.zeros((rad.T.shape[0], rad.T.shape[1], rad.T.shape[1]))
        for idx in range(S_noise.shape[0]):
            S_noise[idx, :, :] = np.diag((meas_radiance['radiance_noise'] / meas_radiance['radiance']).isel(wavelength=idx).values**2) + (meas_norm['radiance_noise'] / meas_norm['radiance']).isel(wavelength=idx).values **2
        S_noise[np.isnan(S_noise)] = 1

        return jacobian.values, S_noise

    def minimize_w(self, K, S, use_constraints=False):

        good_alts = (self._tanalts >= self._min_update_alt - 10) & (self._tanalts <= self._max_alt+10)
        offset = np.where(good_alts)[0][0]

        idx_750 = np.nanargmin(np.abs(np.asarray(self._triplets[0]['wavelengths']) - 750))
        weights = np.zeros((K.shape[0], len(self._tanalts)))
        weights[idx_750] = 1.0

        cons = ({'type': 'eq',
                 'fun': lambda x: np.array(np.linalg.norm(x) - 1)})

        for witer in range(self._weight_iters):
            weights_previous = copy(weights)
            for idx, tanalt in enumerate(self._tanalts[good_alts][::-1]):
                idx = len(self._tanalts[good_alts]) - idx - 1
                if use_constraints:
                    temp = optimize.minimize(self.residual_lognormals_single_dy, x0=weights[:, idx + offset],
                                                args=(K, S, weights, idx, good_alts),
                                                options={'maxiter': 500}, constraints=cons)
                else:
                    temp = optimize.minimize(self.residual_lognormals_single_dy, x0=weights[:, idx + offset],
                                                args=(K, S, weights, idx, good_alts),
                                                options={'maxiter': 500})
                    weights[:, idx + offset] = temp.x / np.linalg.norm(temp.x)

                if (~temp.success) & (witer == self._weight_iters - 1):
                    logging.warning('aerosol weight optimization failed to converge at altitude '
                                    + str(tanalt) + ': ' + temp.message)

            lowest = np.where(good_alts)[0][0]
            weights[:, 0:lowest] = weights[:, lowest][:, np.newaxis]
            highest = np.where(good_alts)[0][-1]
            weights[:, highest:] = weights[:, highest][:, np.newaxis]

            if np.max(np.abs(weights_previous - weights)) < 0.001:
                break
        else:
            logging.warning('aerosol weight optimization not fully converged, difference of: '
                            + str(np.max(np.abs(weights_previous - weights))))

        weights[:, highest:] = weights[:, highest][:, np.newaxis]
        return weights

    @staticmethod
    def residual_lognormals_single_dy(x, K, S, weights, alt_idx, good_alts):
        offset = np.where(good_alts)[0][0]
        weights[:, alt_idx + offset] = x / np.linalg.norm(x)

        try:
            G = np.linalg.inv(np.sum(K[:, good_alts][:, :, good_alts] * weights[:, good_alts, np.newaxis], axis=0))
        except np.linalg.LinAlgError:
            G = np.linalg.pinv(np.sum(K[:, good_alts][:, :, good_alts] * weights[:, good_alts, np.newaxis], axis=0))

        Se = np.sum(S[:, good_alts][:, :, good_alts] * (weights ** 2)[:, good_alts, np.newaxis], axis=0)
        Se = np.diag(np.diag(Se))
        Sx = np.diag(G @ G.T)[alt_idx]

        return Sx

    def rtm_options(self):
        options = dict()
        options['calcwf'] = 2
        options['wfheights'] = self._retrieval_altitudes
        options['wfwidths'] = np.gradient(self._retrieval_altitudes)
        options['numordersofscatter'] = 50
        options['numdiffuseprofilesinplane'] = 1

        return options, 'aerosol'

    def update_state(self, x: np.ndarray):
        current_x = self.state_vector()

        if self._state_vector_space.lower() == 'log':
            deltas = x - current_x
            deltas[deltas < -10] = -10
            deltas[deltas > 10] = 10

            mult_factors = np.exp(deltas)
        else:
            mult_factors = x / current_x

        mult_factors = np.interp(self._atmosphere_altitudes, self._retrieval_altitudes, mult_factors,
                                 left=mult_factors[0], right=mult_factors[-1])

        mult_factors[mult_factors < (1/self._max_update_factor)] = (1/self._max_update_factor)
        mult_factors[mult_factors > self._max_update_factor] = self._max_update_factor

        min_alt_idx = np.nonzero(self._atmosphere_altitudes >= np.nanmax([self._retrieval_altitudes[0], self._min_update_alt]))[0][0]

        mult_factors[0:min_alt_idx+1] = np.exp(-(self._atmosphere_altitudes[min_alt_idx] - self._atmosphere_altitudes[0:min_alt_idx+1])/2000) * (mult_factors[min_alt_idx]-1) + 1

        self._species.climatology[self._species.species] *= mult_factors

    def measurement_vector(self, l1_data: RadianceBase):
        tangent_locations = l1_data.tangent_locations()
        good_data = l1_data.data.copy(deep=True)


        # Filter out below the lowest altitude
        good_data['tangent_altitude'] = tangent_locations['altitude']
        good_data = good_data.swap_dims({'los': 'tangent_altitude'})
        good_data = good_data.where(good_data.tangent_altitude > self._retrieval_altitudes[0] - (self._retrieval_altitudes[1] - self._retrieval_altitudes[0]) / 2, drop=True)

        all_y = []
        all_wf = []
        all_dy = []
        all_y_altitudes = []
        for triplet in self._triplets:
            triplet_data = good_data.sel(wavelength=triplet['wavelengths'], method='nearest')
            good_ray = self._l1_ray.data['radiance'].sel(wavelength=triplet_data.wavelength.values,
                                                         tangent_altitude=triplet_data.tangent_altitude.values,
                                                         method='nearest')
            triplet_data['radiance_ray'] = (['wavelength', 'tangent_altitude'], copy(good_ray.values))

            # Normalize each wavelength first
            triplet_data['log_radiance'] = np.log(triplet_data['radiance'])
            wavel_norm_data = []
            for widx, w in enumerate(triplet_data.wavelength.values):
                norm_range = triplet['wavelength_normalization_range'][widx]
                wavel_norm_data.append(triplet_data.sel(wavelength=w).where((triplet_data.tangent_altitude > norm_range[0]) & (triplet_data.tangent_altitude < norm_range[1]), drop=True).copy(True))

            if 'wf' in wavel_norm_data[0]:
                include_vars = ['radiance', 'radiance_ray', 'relative_wf']
                for data in wavel_norm_data:
                    data['relative_wf'] = data['wf'] / data['radiance']
            else:
                include_vars = ['radiance', 'radiance_ray']

            log_wavel_norm = xr.concat([data[include_vars] for data in wavel_norm_data], dim='wavelength')
            log_wavel_norm['radiance'] = np.log(log_wavel_norm['radiance'])
            log_wavel_norm['radiance_ray'] = np.log(log_wavel_norm['radiance_ray'])
            log_wavel_norm = log_wavel_norm.mean(dim='tangent_altitude')

            normalized_log_triplet_data = triplet_data['log_radiance'] - np.log(triplet_data['radiance_ray']) - log_wavel_norm['radiance'] + log_wavel_norm['radiance_ray']
            normalized_log_triplet_data = normalized_log_triplet_data.transpose('tangent_altitude', 'wavelength')

            # Check to see if our weights are a function of altitude
            if len(np.array(triplet['weights']).shape) == 1:
                triplet_weights = xr.DataArray(triplet['weights'],
                                               coords=[normalized_log_triplet_data.wavelength],
                                               dims=['wavelength'])
            else:
                # Interpolate the weights onto the tangent altitudes
                weights = interpolate.interp1d(triplet['weight_altitudes'], triplet['weights'], axis=1)(normalized_log_triplet_data.tangent_altitude.values)
                triplet_weights = xr.DataArray(weights,
                                               coords=[normalized_log_triplet_data.wavelength, normalized_log_triplet_data.tangent_altitude],
                                               dims=['wavelength', 'tangent_altitude'])

            if not self._require_all_wavelengths_valid:
                # Set nans to 0 and readjust the weights, readjusting the weights doesn't change the retrieval but
                # makes the measurement vector easier to interpret
                nan_idx = np.isnan(normalized_log_triplet_data.values)
                normalized_log_triplet_data.values[nan_idx] = 0

                triplet_weights = triplet_weights.transpose('tangent_altitude', 'wavelength')

                triplet_weights.values[nan_idx] = 0
                triplet_weights.values /= np.linalg.norm(triplet_weights, axis=1)[:, np.newaxis]

            y_triplet = normalized_log_triplet_data.dot(triplet_weights, dims='wavelength')

            if self._high_altitude_normalize and self._double_normalize:
                norm_data = y_triplet.where((y_triplet.tangent_altitude > triplet['normalization_range'][0]) & (y_triplet.tangent_altitude < triplet['normalization_range'][1]))

                y_triplet -= norm_data.mean(dim='tangent_altitude')

            y_triplet = y_triplet.where((y_triplet.tangent_altitude > triplet['lower_altitude_m']) & (y_triplet.tangent_altitude < triplet['upper_altitude_m']), drop=True)

            if 'wf' in good_data:
                wf_data = (triplet_data['wf'] / triplet_data['radiance']).to_dataset(name='wf')
                norm_wf = wf_data.where((wf_data.tangent_altitude > triplet['normalization_range'][0]) & (
                            wf_data.tangent_altitude < triplet['normalization_range'][1]), drop=True)
                wf_data = wf_data.where(
                    (wf_data.tangent_altitude > triplet['lower_altitude_m']) & (
                                wf_data.tangent_altitude < triplet['upper_altitude_m']), drop=True)

                # We might have nans in the wf because of nans in the radiance, but these radiances aren't used
                # so set the values to 0
                wf_data['wf'].values[np.isnan(wf_data['wf'].values)] = 0.0

                if self._high_altitude_normalize_jacobian and self._high_altitude_normalize:
                    if self._double_normalize:
                        wf_data -= norm_wf.mean(dim='tangent_altitude')
                    else:
                        wf_data -= log_wavel_norm['relative_wf']

                if self._state_vector_space == 'log':
                    wf_data['x'] = (['perturbation'], self.state_vector())
                    wf_data['wf'] *= np.exp(wf_data['x'])

                wf = wf_data['wf'].dot(triplet_weights, dims='wavelength')
                all_wf.append(wf.values)

            if 'radiance_noise' in good_data:
                inverse_snr = (triplet_data['radiance_noise'] / np.abs(triplet_data['radiance']))

                if self._assume_error_fraction is not None:
                    inverse_snr.values = np.ones_like(inverse_snr.values) * self._assume_error_fraction

                inverse_snr = inverse_snr.where(
                    (inverse_snr.tangent_altitude > triplet['lower_altitude_m']) & (
                                inverse_snr.tangent_altitude < triplet['upper_altitude_m']), drop=True)

                dy = (inverse_snr**2).dot(triplet_weights**2, dims='wavelength')
                if self._normalization_covariance:
                    # TODO: Fix this for wavelength normalization
                    logging.warning('Normalization covariance does not work for aerosol')
                    # normalization_square_snr = ((normalization_data['radiance_noise'] / normalization_data['radiance'])**2).mean(dim='tangent_altitude')
                    # normalization_triplet_weights = triplet_weights.isel(tangent_altitude=-1)
                    # normalization_dy = normalization_square_snr.dot(normalization_triplet_weights**2, dims='wavelength')
                    #
                    # all_dy.append(np.diag(dy.values) + normalization_dy.values)
                else:
                    all_dy.append(dy.values)

            all_y.append(y_triplet.values)
            all_y_altitudes.append(y_triplet.tangent_altitude.values)
        y_out = dict()
        y_out['y'] = np.hstack(all_y)
        y_out['y_altitude'] = np.hstack(all_y_altitudes)

        if 'wf' in good_data:
            y_out['jacobian'] = np.vstack(all_wf)

        if 'radiance_noise' in good_data:
            if self._normalization_covariance:
                y_out['y_error'] = np.hstack(all_dy)
            else:
                y_out['y_error'] = np.hstack(all_dy).flatten()

        return y_out


    def apriori_state(self):
        if self._use_apriori_for_upperbound:
            if self._state_vector_space == 'log':
                ret_val = np.log(self._initial_aero)
            else:
                ret_val = self._initial_aero

            bounded_altitudes = self._retrieval_altitudes > self._apriori_upperbound
            # This will be all altitudes above the highest altitude, but we need to include one more in order to
            # "link" the actual retrieval altitudes with the upper altitudes
            first_idx = np.nonzero(bounded_altitudes)[0][0]
            bounded_altitudes[first_idx - 1] = True
            ret_val[~bounded_altitudes] = 0

            return ret_val

        else:
            return super().apriori_state()

    def inverse_apriori_covariance(self):
        n = len(self.state_vector())
        use_apriori = False
        inv_Sa = np.zeros((n, n))
        if self._tikh_factor > 0:
            gamma = two_dim_vertical_second_deriv(1, len(self._retrieval_altitudes), self._tikh_factor) / (
                        2 * np.diff(self._retrieval_altitudes)[0])

            if self._use_apriori_for_upperbound:
                bounded_altitudes = self._retrieval_altitudes > self._apriori_upperbound
                first_idx = np.nonzero(bounded_altitudes)[0][0]
                # The altitude just below the upperbound will calculate a derivative using a bounded altitude which
                # we don't want
                bounded_altitudes[first_idx - 1] = True
                bounded_altitudes[first_idx - 2] = True
                gamma[bounded_altitudes, :] = 0

            inv_Sa += gamma.T @ gamma
            use_apriori = True
        if self._apriori_clim is not None:
            if self._state_vector_space == 'log':
                inv_Sa += np.diag(np.ones(n) * (1 / self._apriori_influence ** 2))
            else:
                inv_Sa += np.diag(np.ones(n) * (1 / (self._apriori_influence * self.apriori_state()) ** 2))
            use_apriori = True


        if self._use_apriori_for_upperbound:
            n = len(self.state_vector())

            inv_Sa_upperbound = np.zeros((n, n))
            gamma = two_dim_vertical_first_deriv(1, len(self._retrieval_altitudes), factor=1e4)

            bounded_altitudes = self._retrieval_altitudes > self._apriori_upperbound
            # This will be all altitudes above the highest altitude, but we need to include one more in order to
            # "link" the actual retrieval altitudes with the upper altitudes
            first_idx = np.nonzero(bounded_altitudes)[0][0]
            bounded_altitudes[first_idx - 1] = True
            gamma[~bounded_altitudes, :] = 0

            inv_Sa_upperbound += gamma.T @ gamma

            inv_Sa += inv_Sa_upperbound

        return inv_Sa

    @staticmethod
    def state_vector_allowed_to_change():
        """
        Returns
        -------
        bool
            True if the state vector/apriori may change shape between iterations, False otherwise.
        """
        return True

    @staticmethod
    def measurement_vector_allowed_to_change():
        """
        Returns
        -------
        bool
            True if the measurement_vector may change shape between iterations, False otherwise.
        """
        return True

    def adjust_parameters(self, forward_model, y_dict, chi_sq, chi_sq_linear, iter_idx, predicted_delta_y):
        jacobian_low_alt = self._retrieval_altitudes[0]
        chi_sq_low_alt = self._retrieval_altitudes[0]

        if self._use_jacobian_for_lowerbound:
            jacobian_low_alt = self.min_alt_from_jacobian(y_dict)
            jacobian_low_alt = np.nanmax([jacobian_low_alt, self._min_update_alt])

        if self._use_chi_sq_for_lowerbound:
            y_altitude = y_dict['y_altitude'][~np.isnan(y_dict['y'])]
            # only adjust the lowerbound if we are somewhat close to convergence
            if chi_sq / chi_sq_linear < 5:
                bad_altitudes = y_altitude[np.abs(predicted_delta_y) > 0.002]
                bad_altitudes = bad_altitudes[bad_altitudes < 20000]

                if len(bad_altitudes) > 0:
                    # Remove altitudes below/at the peak of the bad measurements
                    # Removing one altitude at a time is too slow, but removing all of the bad altitudes can be
                    # overly aggressive, so we remove up to and including the peak, or 3 measurements, whichever is lower
                    low_predicted_delta_y = np.abs(predicted_delta_y)[y_altitude < 20000]

                    num_remove = np.nanargmax(low_predicted_delta_y) + 1
                    num_remove = np.min([num_remove, 3])

                    new_lowest_measurement_altitude = y_altitude[num_remove]
                    highest_ret_altitude = self._retrieval_altitudes[self._retrieval_altitudes < new_lowest_measurement_altitude + self._ret_spacing_m / 2][-1]
                    chi_sq_low_alt = highest_ret_altitude

        self.adjust_min_altitude(np.nanmax([chi_sq_low_alt, jacobian_low_alt]))

        forward_model.engine.options['wfheights'] = self._retrieval_altitudes
        forward_model.engine.options['wfwidths'] = np.gradient(self._retrieval_altitudes)