from skretrieval.retrieval import RetrievalTarget
from skretrieval.retrieval.tikhonov import two_dim_vertical_second_deriv
import numpy as np
import sasktran as sk
from skretrieval.core.radianceformat import RadianceGridded, RadianceBase
from abc import abstractmethod


class OsirisTarget(RetrievalTarget):
    @abstractmethod
    def measurement_vector(self, l1_data: RadianceBase):
        pass

    def __init__(self, species: sk.Species, measurement_l1: RadianceGridded, state_vector_space, tikh_factor,
                 lower_altitude_bound_m, upper_altitude_bound_m, ret_spacing_m,
                 apriori_clim, apriori_influence, max_update_factor=2, retrieve_on_osiris_grid=False,
                 allow_negative_values=False, minimum_value=-1e99):
        self._retrieve_on_osiris_grid = retrieve_on_osiris_grid
        if retrieve_on_osiris_grid:
            os_tan_alts = measurement_l1.tangent_locations().altitude.values
            self._retrieval_altitudes = os_tan_alts[os_tan_alts < upper_altitude_bound_m]
            self._retrieval_altitudes = self._retrieval_altitudes[self._retrieval_altitudes > lower_altitude_bound_m]
        else:
            self._retrieval_altitudes = np.arange(lower_altitude_bound_m, upper_altitude_bound_m, ret_spacing_m)

        self._initial_lower_bound = lower_altitude_bound_m
        self._initial_upper_bound = upper_altitude_bound_m
        self._ret_spacing_m = ret_spacing_m

        self._species = species
        self._measurement_l1 = measurement_l1
        self._tikh_factor = tikh_factor

        self._minimum_value = minimum_value

        self._atmosphere_altitudes = self._species.climatology.altitudes

        self._state_vector_space = state_vector_space

        self._apriori_clim = apriori_clim
        self._apriori_influence = apriori_influence
        self._apriori_lat = None
        self._apriori_lon = None
        self._apriori_mjd = None
        self._set_apriori_parameters(measurement_l1)

        self._max_update_factor = max_update_factor

        self._allow_negative_values = allow_negative_values

    def _set_apriori_parameters(self, l1_data: RadianceGridded):
        lats = []
        lons = []
        mjds = []
        mjds.append(np.nanmean(l1_data.data.mjd.values))
        mean_tangent = l1_data.tangent_locations().mean(dim='los')

        lats.append(float(mean_tangent.latitude))
        lons.append(float(mean_tangent.longitude))

        self._apriori_lat = np.nanmean(lats)
        self._apriori_lon = np.nanmean(lons)
        self._apriori_mjd = np.nanmean(mjds)

    def state_vector(self):
        x = self._species.climatology.get_parameter(self._species.species, 0, 0, self._retrieval_altitudes,
                                                        54372)

        if self._state_vector_space.lower() == 'log':
            return np.log(x)
        else:
            return x

    def update_state(self, x: np.ndarray):
        current_x = self.state_vector()

        if self._state_vector_space.lower() == 'log':
            deltas = x - current_x
            deltas[deltas < -10] = -10
            deltas[deltas > 10] = 10

            mult_factors = np.exp(deltas)
        else:
            mult_factors = x / current_x

        # If we are in log space, or we are in linear space and we don't allow negative values
        # then we just restrict the update based on max_update_factor
        if self._state_vector_space == 'log' or not self._allow_negative_values:
            mult_factors[mult_factors < (1/self._max_update_factor)] = (1/self._max_update_factor)
            mult_factors[mult_factors > self._max_update_factor] = self._max_update_factor
        else:
            # We are in linear space and we are allowing negative values
            # We don't want to let the high altitudes go negative however
            if mult_factors[-1] < 1/self._max_update_factor:
                mult_factors[-1] = 1/self._max_update_factor

        mult_factors = np.interp(self._atmosphere_altitudes, self._retrieval_altitudes, mult_factors,
                                 left=mult_factors[0], right=mult_factors[-1])

        self._species.climatology[self._species.species] *= mult_factors
        self._species.climatology[self._species.species][self._species.climatology[self._species.species] < self._minimum_value] = self._minimum_value

    def apriori_state(self):
        if self._apriori_clim is not None:
            x_a = self._apriori_clim.get_parameter('SKCLIMATOLOGY_NO2_CM3', self._apriori_lat, self._apriori_lon,
                                                   self._retrieval_altitudes, self._apriori_mjd)

            if self._state_vector_space == 'log':
                return np.log(x_a)
            else:
                return x_a
        else:
            return None

    def inverse_apriori_covariance(self):
        n = len(self.state_vector())
        use_apriori = False
        inv_S_a = np.zeros((n, n))
        if self._tikh_factor > 0:
            gamma = two_dim_vertical_second_deriv(1, len(self._retrieval_altitudes), self._tikh_factor) / (2*np.diff(self._retrieval_altitudes)[0])
            if self._state_vector_space != 'log':
                inv_S_a += gamma.T @ gamma / np.outer(self.apriori_state(), self.apriori_state())
            else:
                inv_S_a += gamma.T @ gamma
            use_apriori = True
        if self._apriori_clim is not None:
            if self._state_vector_space == 'log':
                inv_S_a += np.diag(np.ones(n)*(1/self._apriori_influence**2))
            else:
                inv_S_a += np.diag(np.ones(n)*(1/(self._apriori_influence * self.apriori_state())**2))
            use_apriori = True

        if use_apriori:
            return inv_S_a
        else:
            return None

    def map_error_covariance(self, covariance):
        # We can only map the error onto altitudes that we actually retrieve on
        map_altitudes_flag = (self._atmosphere_altitudes >= self._retrieval_altitudes[0]) & (self._atmosphere_altitudes <= self._retrieval_altitudes[-1])

        map_altitudes = self._atmosphere_altitudes[map_altitudes_flag]

        def interpolating_matrix(from_grid, to_grid):
            M = np.zeros((len(to_grid), len(from_grid)))

            for idx, ele in enumerate(to_grid):
                try:
                    idx_above = np.nonzero(from_grid > ele)[0][0]
                except:
                    continue

                if idx_above == 0:
                    continue
                else:
                    w = (from_grid[idx_above] - ele) / (from_grid[idx_above] - from_grid[idx_above - 1])
                    M[idx, idx_above] = 1 - w
                    M[idx, idx_above - 1] = w

            return M

        interp_matrix = interpolating_matrix(self._retrieval_altitudes, map_altitudes)
        out_covar = np.ones((len(self._atmosphere_altitudes), len(self._atmosphere_altitudes))) * np.nan

        if self._state_vector_space == 'log':
            x = self.state_vector()
            mapped_covar = interp_matrix @ (np.outer(np.exp(x), np.exp(x)) * covariance) @ interp_matrix.T
        else:
            mapped_covar = interp_matrix @ covariance @ interp_matrix.T

        out_covar[np.ix_(map_altitudes_flag, map_altitudes_flag)] = mapped_covar

        return out_covar
