import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import minimize, minimize_scalar
from skretrieval.util import rotation_matrix
from copy import copy


def min_fn_rsas(x, alt_m, rad_meas, rad_ret):
    rad_meas_shift = np.exp(interp1d(alt_m, np.log(rad_meas), axis=0, bounds_error=False)(alt_m+x))

    low_alt_idx = (alt_m > 18500) & (alt_m < 22500)
    high_alt_idx = (alt_m > 30000) & (alt_m < 36500)

    y = (np.nanmean(rad_meas_shift[low_alt_idx])/np.nanmean(rad_meas_shift[high_alt_idx])) - (np.nanmean(rad_ret[low_alt_idx])/np.nanmean(rad_ret[high_alt_idx]))

    return np.linalg.norm(y)


def rsas_shift(alt_m, rad_meas, rad_ret):
    min_results = minimize(min_fn_rsas, 0, args=(alt_m, rad_meas, rad_ret), method='L-BFGS-B', bounds=[(-4000, 6000)])

    return min_results

def min_fn_rsas_angular(x, l1_meas, l1_ret, wavel_nm):
    meas_shifted = rotate_l1(l1_meas, x)

    rad_meas = meas_shifted.data.sel(wavelength=wavel_nm)['radiance'].values.flatten()
    rad_ret = l1_ret.data.sel(wavelength=wavel_nm)['radiance'].values.flatten()
    alt_m = l1_ret.tangent_locations()['altitude'].values

    rad_meas_shift = interp1d(meas_shifted.tangent_locations()['altitude'], rad_meas, axis=0, bounds_error=False)(alt_m)

    low_alt_idx = (alt_m > 18500) & (alt_m < 22500)
    high_alt_idx = (alt_m > 30000) & (alt_m < 36500)

    y = (np.nanmean(rad_meas_shift[low_alt_idx])/np.nanmean(rad_meas_shift[high_alt_idx])) - (np.nanmean(rad_ret[low_alt_idx])/np.nanmean(rad_ret[high_alt_idx]))

    return np.linalg.norm(y)


def rsas_shift_angular(l1_meas, l1_ret, wavel_nm):
    # min_results = minimize(min_fn_rsas_angular, 0, args=(l1_meas, l1_ret, wavel_nm), method='Nelder-Mead', bounds=[(-0.001, 0.001)])

    min_results = minimize_scalar(min_fn_rsas_angular, 0, args=(l1_meas, l1_ret, wavel_nm), method='bounded', bounds=(-0.001, 0.001))


    return min_results


def rotate_l1(l1_data, angular_rotation_rad):

    new_los = np.zeros_like(l1_data.data.los_vectors.values)

    for idx, los in enumerate(l1_data.data.los.values):
        selected = l1_data.data.sel(los=los)

        rot_axis = np.cross(selected.observer_position.values, selected.los_vectors.values)
        rot_axis /= np.linalg.norm(rot_axis)

        new_los[idx, :] = rotation_matrix(rot_axis, angular_rotation_rad) @ selected.los_vectors.values

    l1_type = type(l1_data)
    new_data = l1_data.data.copy(deep=True)
    new_data.los_vectors.values = new_los

    new_l1 = l1_type(new_data)

    return new_l1