from skretrieval.retrieval import RetrievalTarget
import numpy as np
import sasktran as sk
from skretrieval.core.radianceformat import RadianceBase, RadianceGridded
from copy import copy
from osiris_skretrieval.osirismodel import OSIRISModel
import logging


class BlackCarbonRetrieval(RetrievalTarget):
    def __init__(self, atmosphere: sk.Atmosphere, high_alt: float, forward_model: OSIRISModel, ret_wavel):
        self._atmosphere = atmosphere
        self._ret_wavel = ret_wavel
        self._high_alt = high_alt
        self._ref_range = 5000
        self._numden = 0.3
        self._smallest_numden = 1e-3
        self._highest_numden = 1

        self._clim = atmosphere['blackcarbon'].climatology

        # Set black carbon to 0 to get the meas vector
        # TODO:
        rad = forward_model.calculate_radiance()
        self._zero_meas_vector = None
        self._zero_meas_vector = self.measurement_vector(rad)

        self.update_state(1e-9)

    def measurement_vector(self, l1_data: RadianceBase):
        if not isinstance(l1_data, RadianceGridded):
            raise ValueError('Class BlackCarbonRetrieval only supports data in the form RadianceGridded')

        result = dict()

        tangent_locations = l1_data.tangent_locations()
        l1_data.data['tangent_altitude'] = tangent_locations.altitude

        triplet_values = l1_data.data['radiance'].where((l1_data.data['tangent_altitude'] > self._high_alt - self._ref_range) & (l1_data.data['tangent_altitude'] < self._high_alt), drop=False).sel(wavelength=self._ret_wavel, method='nearest')

        result['y'] = float(np.log(triplet_values.sel(wavelength=self._ret_wavel[0], method='nearest')).mean(dim='los').values) -\
            float(np.log(triplet_values.sel(wavelength=self._ret_wavel[1], method='nearest')).mean(dim='los').values)

        if self._zero_meas_vector:
            result['jacobian'] = (result['y'] - self._zero_meas_vector['y']) / float(self.state_vector())
            result['jacobian'] = result['jacobian'].reshape(1, 1)

        result['y'] = np.array([result['y']])

        return result

    def state_vector(self):
        return np.array([float(self._clim.get_parameter('blackcarbon', 0, 0, 5000, 54372))])

    def update_state(self, x: np.ndarray):
        self._numden = float(x)

        logging.info("Setting black carbon to: {}".format(self._numden))

        self._clim['blackcarbon'] = [self._numden, self._numden]

    def apriori_state(self):
        return None

    def inverse_apriori_covariance(self):
        return None
